' **
' * Title: FAT16Driver.bas
' * Date: 02/08/03
' * Author: Larry Wolcott
' * email: objectcoder@hotmail.com
' * web: http://www.larrywolcott.com
' *
' * Description: A functional reference implementation of FAT16. Requires
' * FlashCardDrive.bas by Josef Franz Vogel found at http://www.mcselec.com 
in the
' * Application Notes area listed as AN-123. A simple CompactFlash hardware
' * inteface is also required. See additional documentation enclosed or 
refer
' * to the "Accessing a Compact Flash Card from BASCOM" (AN-123) on the 
site.
' * see http://www.mcselec.com/an_123.htm
' *
' * This has been developed for the Atmega103/128 AVRs. Memory requirements
' * are such that this driver should work with any AVR that has 2K SRAM or 
more.
' * Furthermore, there is currently no cluster chaninig support other than
' * basic, linear linking. That means the code will not reuse space created
' * by deleting files. When the storage device reaches maximun capacity, it
' * should be formated. This is the safest way to ensure cross-linked 
cluster
' * chains do not occur.
' *
' * NOTE: Be sure to set the 'Frame Size' parameter to 256 or this driver 
will
' * not work properly.
' **

'include the low-level comms driver for the flash card (AN-123 by Josef 
Franz)
$Include "FlashCardDrive.bas"

'This function is used to return a long datatype from the buffer
Declare Function Getlongfrombuffer(pbsramarray As Byte , Byval Pbpos As 
Word) As Long

'This function is used to return a word datatype from the buffer
Declare Function Getwordfrombuffer(pbsramarray As Byte , Byval Pbpos As 
Word) As Word

'This function is used to read the Master Boot Record
Declare Sub Readmbr()

'This function is used init the storage device
Declare Sub Init()

'used to convert a long value to 4 bytes
Declare Sub Convertlong(byval Longvalue As Long)

'used to convert a word value to 2 bytes
Declare Sub Convertword(byval Wordvalue As Word)

'this sub prints a message to the RS232 or LCD if the debug is set
Declare Sub Debug(byval Debugtext As String)

'this sub sets the debug output device
Declare Sub Setdebug(byval Debugvalue As Byte)

'this sub is used to read boot record for a given partition
Declare Sub Readbootrecord()

'this function is used to parse a 512 byte sector for the 32 byte fat record
Declare Function Parsefatrecord(byval Recordnum As Word) As Byte

'this function is used to search the root dir for a file
Declare Function Searchrootdir(byval Searchname As String) As Byte

'this function is used to calculate the sector from the cluster
Declare Function Findsector(byval Clusternumber As Word) As Long

'checks to see if a file exists, if not, it gets created
Declare Sub Createorappendtofile(byval Fname As String)

'finds the sector and entry number on the root that is available
Declare Sub Pointtofreeentry()

'creates a new directory entry on the root
Declare Sub Createrootentry(byval Touchname As String)

'writes data to the storage device
Declare Function Writedata(byval Texttowrite As String , Byval Linefeed As 
Byte) As Byte

'calculates the next cluster/sector to use for a new file
Declare Sub Findnextavailable()

'updates the allocation table
Declare Sub Updateallocationtable(byval Isnew As Byte)

'loads the databuffer
Declare Sub Loaddatabuffer()

'this sub is used to list the files on the root dir
Declare Sub Listrootdir()

'sets the dir values to a record on a speficied sector
Declare Sub Setdirvalues()

'finds the next available cluster
Declare Sub Findlastusedsector()

'This function is used to dump an arbitrary sector as hex
Declare Sub Dumpsectorashex(byval Dumpsect As Long)

'This function is used to dump an arbitrary sector as chr
Declare Sub Dumpsectoraschr(byval Dumpsect As Long)

'lists a file contents as ascii text
Declare Sub Listfile(byval Fname As String)

'This function is used to read the Master Boot Record
Declare Sub Dumpfile(byval Startcluster As Word , Byval Filelen As Long)

'values used by the hex conversion routines
Dim Conv_long As Long
Dim Conv_string As String * 8
Dim Conv_bytes(4) As Byte
Dim Conv_strbytes(4) As String * 2
Dim Convw_word As Word
Dim Convw_string As String * 4
Dim Convw_bytes(2) As Byte
Dim Convw_strbytes(2) As String * 2

'master boot record (MBR) values of interest
Dim Mbr_fstype As Byte
Dim Mbr_sectfrom As Long
Dim Mbr_fatoffset As Long
Dim Mbr_sectorsinroot As Byte
Dim Mbr_totalsectors As Long

'boot record values of interest
Dim Br_bytespersector As Word
Dim Br_sectorspercluster As Byte
Dim Br_reservedsectors As Word
Dim Br_copiesoffat As Byte
Dim Br_sectorsperfat As Word
Dim Br_rootcluster As Long
Dim Br_maxrootentries As Word
Dim Br_sectorsinroot As Word
Dim Br_actualstartsector As Long
Dim Br_sectorsinallfats As Long

'define the standard FAT entry fields
Dim Fat_shortname(8) As Byte
Dim Fat_extension(3) As Byte
Dim Fat_attribute As Byte
Dim Fat_createdms As Byte
Dim Fat_createdtm As Word
Dim Fat_createddt As Word
Dim Fat_accessdt As Word
Dim Fat_fathighcluster As Word
Dim Fat_updatetime As Word
Dim Fat_updatedate As Word
Dim Fat_cluster As Word
Dim Fat_filesize As Long
Dim Fat_volumename As String * 11
Dim Fat_filename As String * 12
Dim Fat_dirname As String * 12
Dim Fat_direntry As Word
Dim Fat_dirsector As Long

'vlues for the update allocation table routine
Dim Alloc_currentcluster As Word
Dim Alloc_currentsector As Long
Dim Alloc_lastcluster As Word
Dim Allocpointer As Word
Dim Alloc_clustoffset As Word
Dim Alloc_clustcalc1 As Long
Dim Alloc_clustcalc2 As Long
Dim Alloc_cntr As Byte
Dim Alloc_temp1 As Long
Dim Alloc_isnewsector As Byte

'values for the last used sector routing
Dim Lus_cluster As Word
Dim Lus_prevcluster As Word
Dim Lus_sector As Long
Dim Lus_fatsector As Long
Dim Lus_cntr As Long
Dim Lus_rec As Word
Dim Lus_loop As Word
Dim Lus_bytes As Long
Dim Lus_sectorincluster As Byte
Dim Lus_isnewfile As Byte
Dim Lus_calc1 As Long
Dim Lus_calc2 As Long
Dim Lus_sectoroffset As Word

'values for the write routine
Dim Wr_len As Word
Dim Wr_text As String * 128
Dim Wr_text2 As String * 128
Dim Wr_calc1 As Long
Dim Wr_calc2 As Long
Dim Wr_calc3 As Long
Dim Wr_calc4 As Long
Dim Wr_cntr As Word
Dim Wr_char As String * 1
Dim Wr_sectloop As Byte
Dim Wr_sectcount As Byte

'values for the dump file routine
Dim Df_totalsecinfile As Long
Dim Df_secinfilebytes As Long
Dim Df_sectcount As Byte
Dim Df_modbytesinfile As Word
Dim Df_bytestoread As Word
Dim Df_cntr As Long
Dim Df_cntr2 As Word
Dim Df_startcluster As Long
Dim Df_startsector As Long
Dim Df_length As Long

'various global values
Dim G_debug As Byte
Dim G_debugtext As String * 80
Dim G_offset As Word
Dim G_cntr As Long
Dim G_rootloop As Byte
Dim G_pointer As Word
Dim G_datapointer As Word
Dim G_sectnum As Long
Dim G_sectcount As Byte
Dim G_parse_temp As Long
Dim G_attribute As Byte
Dim G_attribreturn As Byte
Dim G_temp_cntr As Byte
Dim G_temp_val As Long
Dim G_volumerecord As Word
Dim G_touchvalue As Byte
Dim G_touchholder As String * 12
Dim G_dotposition As Byte
Dim G_dotloop As Byte
Dim G_dotextcount As Byte
Dim G_dotchar As String * 1

'values for the find sector routine
Dim Fnd_sectorcalc As Long
Dim Fnd_spcoffset As Long
Dim Set_diroffset As Word

'data buffer values
Dim Data_sectoffset As Long
Dim Data_sector As Long
Dim Data_sectorcluster As Long

'global 512 byte transfer buffers
Dim G_xferbuffer(512) As Byte
Dim G_databuffer(512) As Byte
Dim G_allocbuffer(512) As Byte

' **
' * This function is used init the storage device
' **
Sub Init()

  Print "Init()"

  'first, read the MBR (Master Boot Record)
  Call Readmbr()

  'second, read the boot record of the first partition
  Call Readbootrecord()

  'find the last used cluster/sector in the dir entry
  Call Findlastusedsector()

End Sub

' **
' * This function is used to read the Master Boot Record
' **
Sub Dumpfile(byval Startcluster As Word , Byval Filelen As Long)

  Df_length = Filelen
  Df_startsector = Findsector(startcluster)

  Print "Dumping File, sect: " ; Df_startsector ; " bytes: " ; Df_length

  'if the file is larger than a single sector, calculate how many
  If Df_length > 512 Then

    'figure out how man 512 byte sectors and the mod bytes
    Df_totalsecinfile = Df_length / 512
    Df_secinfilebytes = Df_totalsecinfile * 512
    Df_modbytesinfile = Df_length - Df_secinfilebytes
  Else

    'the file fits on a single sector
    Df_totalsecinfile = 0
    Df_modbytesinfile = Df_length
  End If

  Print "Total sec: " ; Df_totalsecinfile ; "  mod: " ; Df_modbytesinfile ; 
" bytes"

  'point to the first byte of the array
  G_datapointer = Varptr(g_databuffer(1))

  'MBR fits on a simgle 512 byte sector
  Df_sectcount = 1
  Df_totalsecinfile = Df_totalsecinfile + 1
  Df_bytestoread = 512

  'for each of the sectors
  For Df_cntr = 1 To Df_totalsecinfile

    If Df_cntr = Df_totalsecinfile Then
      Df_bytestoread = Df_modbytesinfile
    End If

    'read sector number 1 into the 512 byte array starting at the pointer 
address
    Readsector G_datapointer , Df_startsector , Df_sectcount

    For Df_cntr2 = 1 To Df_bytestoread
      Print Chr(g_databuffer(df_cntr2)) ;
    Next Df_cntr2

    Df_startsector = Df_startsector + 1

  Next Df_cntr
  Print ""
  Print "*EOF*"

End Sub

' **
' * This function is used to dump an arbitrary sector as chr
' **
Sub Dumpsectoraschr(byval Dumpsect As Long)

  Print "Dumping Sector: " ; Dumpsect

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  'read sector number 1 into the 512 byte array starting at the pointer 
address
  Readsector G_pointer , Dumpsect , 1

  For G_cntr = 1 To 512
    Print Chr(g_xferbuffer(g_cntr)) ;
  Next G_cntr
  Print ""
  Print "*EOF*"

End Sub

' **
' * This function is used to dump an arbitrary sector as hex
' **
Sub Dumpsectorashex(byval Dumpsect As Long)

  Print "Dumping Sector: " ; Dumpsect

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  'read sector number 1 into the 512 byte array starting at the pointer 
address
  Readsector G_pointer , Dumpsect , 1

  For G_cntr = 1 To 512
    Print Hex(g_xferbuffer(g_cntr)) ; " " ;
  Next G_cntr
  Print ""
  Print "*EOF*"
End Sub

' **
' * Writes data to the storage device
' **
Function Writedata(byval Texttowrite As String , Byval Linefeed As Byte) As 
Byte

  'Set The Value Outside Of The Current Frame
  Wr_text = Texttowrite

  Wr_len = Len(wr_text)
  Print "Writing " ; Wr_len ; " bytes: '" ; Wr_text ; "'"

  If Linefeed > 0 Then
   Wr_text = Wr_text + Chr(13)
   Wr_text = Wr_text + Chr(10)
   Print "Adding 2 byte linefeed to text"
  End If

  'Point To The First Byte Of The Array
  G_pointer = Varptr(g_xferbuffer(1))

  'Set The Filesize To + Length Of Bytes
  Wr_len = Len(wr_text)

  'update the fat file size
  Fat_filesize = Fat_filesize + Wr_len

  Print "Loading data buffer to write"

  'update the data buffer
  Call Loaddatabuffer()

  Print "Setting the directory entry values"

  'set the dir values to the xfer buffer before writing
  Call Setdirvalues()

  Print "Writing the directory entry values"

  'write the xferbuffer to the appropriate sector on the storage media
  Writesector G_pointer , Fat_dirsector , 1

  'extend the cluster chain if required
  If Lus_sectorincluster > Br_sectorspercluster Then
    'save the previous cluster number
    Lus_prevcluster = Lus_cluster

    'update the last used cluster
    Lus_cluster = Lus_cluster + 1

    'point th the first sector in the cluster
    Lus_sectorincluster = 1

    Print "UPDATE ALLOC, prev:" ; Lus_prevcluster ; " new:" ; Lus_cluster

    'update the allocation table
    Call Updateallocationtable(1)
  End If

End Function

' **
' * loads the databuffer
' **
Sub Loaddatabuffer()

  'Print "Loading data buffer"

  'reset the calculator value
  Wr_calc2 = 0
  Wr_text2 = ""

  'Point To The First Byte Of The Array
  G_datapointer = Varptr(g_databuffer(1))

  'add the test length to the current sector offset
  Wr_calc1 = Lus_sectoroffset + Wr_len

  'see if the total exceeds a single sector
  If Wr_calc1 > 512 Then
    Wr_sectcount = 2
    Wr_calc3 = Wr_calc1 - 512
    Wr_calc4 = Wr_len - Wr_calc3
  Else
    Wr_calc3 = 0
    Wr_sectcount = 1
  End If

  'for each sector we are writing. 1 or 2, really
  For Wr_sectloop = 1 To Wr_sectcount

    'read sector number 1 into the 512 byte array starting at the pointer 
address
    Readsector G_datapointer , Lus_sector , 1

    'Print "Chars: '" ;

    'write the buytes
    For Wr_cntr = 1 To Wr_len
      'inc the calc
      'Wr_calc2 = Lus_sectoroffset + Wr_cntr
      Lus_sectoroffset = Lus_sectoroffset + 1

      'get the character from the string position as a byte
      Wr_char = Mid(wr_text , Wr_cntr , 1)

      If Lus_sectoroffset < 513 Then
        'set the bytes in the buffer
        G_databuffer(lus_sectoroffset) = Wr_char
        'Print Wr_char ;
      Else
        Wr_text2 = Wr_text2 + Wr_char
      End If

    Next Wr_cntr

    'Print "'"
    Print "Writing sector " ; Lus_sector ; " sectinclust: " ; 
Lus_sectorincluster

    'write the xferbuffer to the appropriate setcor on the storage media
    Writesector G_datapointer , Lus_sector , 1

    'if there is another sector, set the values for a second pass
    If Wr_sectcount > 1 Then
      If Wr_sectloop < 2 Then
        Lus_sector = Lus_sector + 1
        Lus_sectorincluster = Lus_sectorincluster + 1
        Wr_text = Wr_text2
        Wr_len = Len(wr_text)
        Lus_sectoroffset = 0
      End If
    End If


    'Print "new sector offset: " ; Lus_sectoroffset

  Next Wr_sectloop

End Sub

' **
' * calculates the next cluster/sector to use for a new file
' **
Sub Findnextavailable()

  Lus_sectoroffset = 0

  'if empty, then set to the min
  If Lus_cluster < 2 Then

    Print "First cluster entry- setting to 2"

    Lus_cluster = 2
    Lus_bytes = 0
    Lus_sector = Findsector(lus_cluster)
    Lus_sectorincluster = 1
  Else

    'if this is not a new file
    If Lus_isnewfile = 0 Then
      Print "Calculating cluster/sector/byte to append"

      If Fat_filesize > 512 Then
        Lus_calc1 = Fat_filesize / 512
        Lus_calc2 = Lus_calc1 / Br_sectorspercluster
        Lus_sectorincluster = Lus_calc1 Mod Br_sectorspercluster
        Lus_sectorincluster = Lus_sectorincluster + 1
        Lus_sectoroffset = Fat_filesize Mod 512

        Lus_sector = Findsector(fat_cluster)
        Lus_sector = Lus_sector + Lus_calc1
        Lus_cluster = Lus_cluster + Lus_calc2
      Else
        Lus_sectoroffset = Fat_filesize Mod 512
        Lus_sectorincluster = 1
        Lus_sector = Findsector(fat_cluster)
      End If

    'this is a new file
    Else
      Print "Calculating next cluster/sector/byte for new file"
      Lus_sectoroffset = 0
      Lus_sectorincluster = 1

      'the previous file was greater than 1 sector
      If Lus_bytes > 512 Then
        Lus_calc1 = Lus_bytes / 512
        Lus_calc1 = Lus_calc1 + 1                           'add one for the 
mod

        'the previous file fit in to a silgle cluster
        If Lus_calc1 <= Br_sectorspercluster Then
          Print "previous file was greater than 512 but less than a cluster"
          Lus_cluster = Lus_cluster + 1
        Else
          Print "previous file was greater than 512 and more than a cluster"
          Lus_calc2 = Lus_calc1 / Br_sectorspercluster
          Lus_calc2 = Lus_calc2 + 1
          Lus_cluster = Lus_cluster + Lus_calc2
        End If

        Lus_sector = Findsector(lus_cluster)
        Lus_sectorincluster = 1

      'previous file was less than one sector so te calc is easy
      Else
        Print "previous file was less then 1 sector (512 bytes)"
        Lus_cluster = Lus_cluster + 1
        Lus_sector = Findsector(lus_cluster)
        Lus_sectorincluster = 1
      End If

    End If

    Print "Use clust: " ; Lus_cluster ; " sect: " ; Lus_sector ; " 
secinclus: " ;
    Print Lus_sectorincluster ; " secoffset: " ; Lus_sectoroffset
  End If

End Sub

' **
' * checks to see if a file exists, if not, it gets created
' **
Sub Createorappendtofile(byval Fname As String)

  'search the root dir for the datafile
  G_touchvalue = Searchrootdir(fname)

  'check the search results
  If G_touchvalue > 0 Then

    Print "File: " ; Fname ; " clust: " ;
    Print Fat_cluster ; " size: " ; Fat_filesize
    Lus_isnewfile = 0

    'find the cluster/sector/byte offset for the existing file
    Call Findnextavailable()

  'file not found so create it
  Else

    Print "File: " ; Fname ; " not found "
    Lus_isnewfile = 1

    'calculate the next available cluster
    Call Findnextavailable()

    'create a root entry for the file
    Call Createrootentry(fname)

    'update the allocation table
    Call Updateallocationtable(0)

  End If

End Sub

' **
' * updates the allocation table
' **
Sub Updateallocationtable(byval Isnew As Byte)

  'flag to track if this entry spans two sectors
  Alloc_isnewsector = 0

  'point to the first byte of the array
  Allocpointer = Varptr(g_allocbuffer(1))

  'calculate the cluster offset
  Alloc_clustcalc1 = Lus_cluster * Mbr_fatoffset
  Alloc_clustcalc1 = Alloc_clustcalc1 + 1

  'how many sectors
  Alloc_clustoffset = Alloc_clustcalc1 / 512
  Alloc_clustcalc1 = Alloc_clustcalc1 Mod 512

  'calculate what sector to read for the alloction table
  Alloc_currentsector = Mbr_sectfrom + Alloc_clustoffset
  Alloc_currentsector = Alloc_currentsector + 1

  Print "Reading allocation table at sector: " ; Alloc_currentsector

  'read sector number 1 into the 512 byte array starting at the pointer 
address
  Readsector Allocpointer , Alloc_currentsector , 1

  'update the bytes
  Print "setting " ; Mbr_fatoffset ; " bytes at " ; Alloc_clustcalc1

  'this is part of an existing cluster chain so set the previous entry to 
here
  If Isnew = 1 Then

    If Alloc_clustcalc1 > 2 Then

      'back up two bytes
      Alloc_clustcalc1 = Alloc_clustcalc1 - 2

      'convert the previous cluster 'word' to two bytes
      Call Convertword(lus_cluster)

      'set the two byte words
      G_allocbuffer(alloc_clustcalc1) = Convw_bytes(1)
      Alloc_clustcalc1 = Alloc_clustcalc1 + 1
      G_allocbuffer(alloc_clustcalc1) = Convw_bytes(2)
      Alloc_clustcalc1 = Alloc_clustcalc1 + 1

    Else

      'indscates that this entry will span two sectors
      Alloc_isnewsector = 1

    End If
  End If

  'set the 'end of chain' marker
  G_allocbuffer(alloc_clustcalc1) = &HFF
  Alloc_clustcalc1 = Alloc_clustcalc1 + 1
  G_allocbuffer(alloc_clustcalc1) = &HFF

  'write the xferbuffer to the appropriate setcor on the storage media
  Writesector Allocpointer , Alloc_currentsector , 1

  'we man need to back up one sector because the chain spans over to a new 
one
  If Alloc_isnewsector = 1 Then

    'read sector into the 512 byte array starting at the pointer address
    Readsector Allocpointer , Alloc_temp1 , 1

    'convert the previous cluster 'word' to two bytes
    Call Convertword(lus_cluster)

    'set the two byte words
    G_allocbuffer(511) = Convw_bytes(1)
    G_allocbuffer(512) = Convw_bytes(2)

    'debug info
    Print "Writing allocation table at sector: " ; Alloc_currentsector

    'write sector from the 512 byte array starting at the pointer address
    Writesector Allocpointer , Alloc_temp1 , 1

  End If

End Sub

' **
' * finds the sector and entry number on the root that is available
' **
Sub Pointtofreeentry()

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  'start by adding the starting sector with the reserved sectors
  Br_actualstartsector = Mbr_sectfrom + Br_reservedsectors

  'calculate all of the sectors in every copy of the FAT table
  Br_sectorsinallfats = Br_sectorsperfat * Br_copiesoffat

  'add the total sectors in all copies of the FAT table
  Br_actualstartsector = Br_actualstartsector + Br_sectorsinallfats

  G_rootloop = 0

  While G_rootloop <= Br_sectorsinroot

     'read sector number 1 into the 512 byte array starting at the pointer 
address
     Readsector G_pointer , Br_actualstartsector , 1

     'set the fat directory sector for future updates
     Fat_dirsector = Br_actualstartsector

     'start with record 0
     G_volumerecord = 0

     'examine each of 16 records on this sector until the volume attribute 
is found
     While G_volumerecord < 16

       'set the fat directory entry number for future updates
       Fat_direntry = G_volumerecord

       G_attribreturn = Parsefatrecord(g_volumerecord)

         'the attribute indicates this entry is a file
         If G_attribreturn = 0 Then
            Print "Empty dir entry at location " ; Fat_direntry ; " on 
sector " ; Fat_dirsector
            G_volumerecord = 254
            G_rootloop = 254
         End If

         'increase the volume record counter
       G_volumerecord = G_volumerecord + 1

     Wend

     'increase the sector and loop counters
     Br_actualstartsector = Br_actualstartsector + 1
     G_rootloop = G_rootloop + 1

  Wend
End Sub

' **
' * creates a new directory entry on the root
' **
Sub Createrootentry(byval Touchname As String)

  'copy the frame variable to the stack
  G_touchholder = Touchname

  'set the dot position to zero
  G_dotposition = 0
  G_dotextcount = 0

  'point the buffer to an empty entry
  Call Pointtofreeentry()

  'iterate the filename characters and copy to the shortname and extension
  For G_dotloop = 1 To Len(g_touchholder)

    'get the single character from the string position
    G_dotchar = Mid(g_touchholder , G_dotloop , 1)

    'if its the left side of the dot, then build the shortname
    If G_dotposition = 0 Then

      'if dot was found, skip and mark it
      If G_dotchar = "." Then
        'Print "." ;
        G_dotposition = G_dotloop
      Else
        'set the shortname value
        Fat_shortname(g_dotloop) = G_dotchar
        'Print Dotchar ;
      End If

    'this must be the right side of the dot
    Else

     'set the byte position of the extension and assign a value
     G_dotextcount = G_dotextcount + 1
     Fat_extension(g_dotextcount) = G_dotchar
     'Print Dotchar ;
    End If
  Next Dotloop

  'set the various field values to write
  Fat_attribute = 32
  Fat_createdms = 0
  Fat_createdtm = 0
  Fat_createddt = 0
  Fat_accessdt = 0
  Fat_fathighcluster = 0
  Fat_updatetime = 0
  Fat_updatedate = 0
  Fat_cluster = Lus_cluster
  Fat_filesize = 0

  'set the dir values to the xfer buffer before writing
  Call Setdirvalues()

  'write the xferbuffer to the appropriate setcor on the storage media
  Writesector G_pointer , Fat_dirsector , 1

End Sub

' **
' * sets the dir values to a record on a speficied sector
' **
Sub Setdirvalues()

  'byte offset for the directory entry
  Set_diroffset = Fat_direntry * 32

  'set the xferbuffer bytes for the shortname
  For G_cntr = 1 To 8
    Set_diroffset = Set_diroffset + 1

    If Fat_shortname(g_cntr) > 0 Then
      G_xferbuffer(set_diroffset) = Fat_shortname(g_cntr)
    Else
      G_xferbuffer(set_diroffset) = 32
    End If
  Next G_cntr

  'set the xferbuffer bytes for the extension
  For G_cntr = 1 To 3
    Set_diroffset = Set_diroffset + 1

    If Fat_extension(g_cntr) > 0 Then
      G_xferbuffer(set_diroffset) = Fat_extension(g_cntr)
    Else
      G_xferbuffer(set_diroffset) = 32
    End If
  Next G_cntr

  'set the attribute
  Set_diroffset = Set_diroffset + 1
  G_xferbuffer(set_diroffset) = Fat_attribute

  'skip the reserved byte
  Set_diroffset = Set_diroffset + 1

  'set the rest of the values
  'For Cntr = 1 To 13
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H0C
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H20
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H91
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H3C
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H2E
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H3C
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H2E
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H00
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H00
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H28
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H91
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H3C
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = &H2E

  'Next Cntr

  Call Convertword(fat_cluster)

  'set the rest of the values
  For G_cntr = 1 To 2
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = Convw_bytes(g_cntr)
    'Print Convw_bytes(cntr) ; " " ;
  Next G_cntr

  'Print "  =  " ;

  Call Convertlong(fat_filesize)

  'set the rest of the values
  For G_cntr = 1 To 4
    Set_diroffset = Set_diroffset + 1
    G_xferbuffer(set_diroffset) = Conv_bytes(g_cntr)
    'Print Conv_bytes(cntr) ; " " ;
  Next G_cntr

End Sub

' **
' * finds the next available cluster
' **
Sub Findlastusedsector()

  Print "FindLastUsedSector()"

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  Lus_fatsector = Br_actualstartsector
  Lus_sector = 0

  Print "Finding Last Used Cluster starting at: " ; Lus_fatsector

  'MBR fits on a simgle 512 byte sector
  Lus_loop = 0

  While Lus_loop <= Br_sectorsinroot

     'read sector number 1 into the 512 byte array starting at the pointer 
address
     Readsector G_pointer , Lus_fatsector , 1

     'start with record 0
     Lus_rec = 0

     'exame each of 16 records on this sector until the volume attribute is 
found
     While Lus_rec < 16

         'parse the record
         G_attribreturn = Parsefatrecord(lus_rec)

         'the attribute indicates this entry is a file
         If G_attribreturn > 31 And Fat_shortname(1) <> 229 Then

           Print "file: " ; Fat_filename ; " at cluster: " ; Fat_cluster ;
           Print " size: " ; Fat_filesize

           If Fat_cluster > Lus_cluster Then
             Print "Setting last used cluster to: " ; Fat_cluster
             Lus_cluster = Fat_cluster
             Lus_sector = Findsector(lus_cluster)
             Lus_bytes = Fat_filesize
           End If

         End If

       Lus_rec = Lus_rec + 1
     Wend

     Lus_fatsector = Lus_fatsector + 1
     Lus_loop = Lus_loop + 1
  Wend
End Sub

' **
' * This function is used to calculate the sector from the cluster
' **
Function Findsector(byval Clusternumber As Word) As Long

  'calc the sectors based on the cluster * sectors per cluster
  Fnd_sectorcalc = Br_sectorspercluster * Clusternumber

  Fnd_spcoffset = Br_sectorspercluster * 2

  Fnd_sectorcalc = Fnd_sectorcalc - Fnd_spcoffset

  'cluster
  Findsector = Data_sector + Fnd_sectorcalc

End Function

' **
' * This function is used to read the Master Boot Record of the storage 
device.
' * The resulting values are stored in the global variables starting with 
Mbr_
' **
Sub Readmbr()

  'output some debug info
  Call Debug( "ReadMBR()")

  'set the offset values
  G_offset = 447

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  'MBR is always sector #0
  G_sectnum = 0

  'debug
  Print "Reading MBR at Sector: " ; G_sectnum
  'Call Debug(g_debugtext)

  'read sector number 1 into the 512 byte array starting at the pointer 
address
  Readsector G_pointer , G_sectnum , 1

  'see if the value is greater than 0, otherwise, that partition not defined
  If G_xferbuffer(g_offset) > 0 Then

     'update the offset value
     G_offset = G_offset + 4

     'read the filesystem type from the transfer buffer
     Mbr_fstype = G_xferbuffer(g_offset)
     Print "File System Type: " ; Hex(mbr_fstype)

     G_offset = G_offset + 3

     'read the starting sector for the partition
     Mbr_sectfrom = Getlongfrombuffer(g_xferbuffer(1) , G_offset)
     Print "Start Sector: " ; Mbr_sectfrom

     If Mbr_fstype = &H0B Then
       Mbr_fatoffset = 4
       Mbr_sectorsinroot = 0
       Print "Filesystem: FAT32"
     End If

     If Mbr_fstype = &H06 Then
       Mbr_fatoffset = 2
       Mbr_sectorsinroot = 14
       Print "Filesystem: FAT16"
     End If

     If Mbr_fstype < &H06 Then
       Mbr_fatoffset = 1.5
       Mbr_sectorsinroot = 32
       Print "Filesystem: FAT12"
     End If

  End If

End Sub

' **
' * This function is used to read boot record for a given partition
' **
Sub Readbootrecord()

  Print "ReadBootRecord()"

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  Print "Reading from Sector " ; Mbr_sectfrom

  'find the starting sector from the partition table lookup
  G_sectnum = Mbr_sectfrom

  'read sector number 1 into the 512 byte array starting at the pointer 
address
  Readsector G_pointer , G_sectnum , 1

  'set the offset and get the bytes per sector field
  G_offset = 11
  Br_bytespersector = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

  'set the offset and get the sectors per cluster field
  G_offset = G_offset + 2
  Br_sectorspercluster = G_xferbuffer(g_offset + 1)

  'set the offset and get the reserved sectors
  G_offset = G_offset + 1
  Br_reservedsectors = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

  'set the offset and get the copies of fat
  G_offset = G_offset + 2
  Br_copiesoffat = G_xferbuffer(g_offset + 1)

  'set the offset and get the max number of root entries
  G_offset = G_offset + 1
  Br_maxrootentries = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

  'set the offset and get the sectors per fat
  If Mbr_fstype < &H0B Then

    'reads the sectors per fat for FAT12/16
    G_offset = G_offset + 5
    Br_sectorsperfat = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

  Else
    'reads the sectors per fat for FAT32
    G_offset = G_offset + 19
    Br_sectorsperfat = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'set the offset and get the root directory starting cluster
    G_offset = G_offset + 1
    Br_rootcluster = Getlongfrombuffer(g_xferbuffer(1) , G_offset)

  End If

  'start by adding the starting sector with the reserved sectors
  Br_actualstartsector = Mbr_sectfrom + Br_reservedsectors

  'calculate all of the sectors in every copy of the FAT table
  Br_sectorsinallfats = Br_sectorsperfat * Br_copiesoffat

  'add the total sectors in all copies of the FAT table
  Br_actualstartsector = Br_actualstartsector + Br_sectorsinallfats

  'calc the data sector offset based on the max entries * 32 (32 bytes each)
  Data_sectoffset = Br_maxrootentries * 32

  'devide the offset by bytes per sector to determine the sector number
  Data_sectorcluster = Data_sectoffset / Br_bytespersector

  Data_sector = Br_actualstartsector + Data_sectorcluster

  'For Cntr = 1 To 120
  ' Print Xferbuffer(cntr) ; " ";
  'Next Cntr
  'Print " "

  Print "Bytes per sector: " ; Br_bytespersector
  Print "Sectors per cluster: " ; Br_sectorspercluster
  Print "Sectors per fat: " ; Br_sectorsperfat
  Print "Reserved sectors: " ; Br_reservedsectors
  Print "Copies of fat: " ; Br_copiesoffat
  Print "Max root entries: " ; Br_maxrootentries

End Sub

' **
' * This function is used to return a long datatype from the buffer
' *
' * Code originated by Josef Franz
' **
Function Getlongfrombuffer(pbsramarray As Byte , Byval Pbpos As Word) As 
Long
   ' Extract a Long-Value from a Byte-Array
   ' pbSRAMArray: Byte-array, from which the Long-value should be extracted
   ' pbPos: Position, at which the Long-Value starts (0-based)
   Loadadr Pbsramarray , Z
   Loadadr Pbpos , X
   ld r24, x+
   ld r25, x+
   add zl, r24
   adc zh, r25
   Loadadr Getlongfrombuffer , X
   ldi r24, 4
   !Call _Mem_Copy
End Function

' **
' * This function is used to return a word datatype from the buffer
' *
' * Code originated by Josef Franz
' **
Function Getwordfrombuffer(pbsramarray As Byte , Byval Pbpos As Word) As 
Word
   ' Extract a Word-value from a Byte-Array
   ' pbSRAMArray: Byte-array, from which the Word-value should be extracted
   ' pbPos: Position, at which the Word-Value starts (0-based)
   Loadadr Pbsramarray , Z
   Loadadr Pbpos , X
   ld r24, x+
   ld r25, x+
   add zl, r24
   adc zh, r25
   Loadadr Getwordfrombuffer , X
   ldi r24, 2
   !Call _Mem_Copy
End Function

' **
' * Used to convert a long value to 4 bytes
' **
Sub Convertlong(byval Longvalue As Long)

  'set the long value outside the frame
  Conv_long = Longvalue
  Conv_string = Hex(conv_long)

  'parse the HEX string value as 2 char strings
  Conv_strbytes(1) = Mid(conv_string , 1 , 2)
  Conv_strbytes(2) = Mid(conv_string , 3 , 2)
  Conv_strbytes(3) = Mid(conv_string , 5 , 2)
  Conv_strbytes(4) = Mid(conv_string , 7 , 2)

  'convert the bit order to little endian
  Conv_bytes(4) = Hexval(conv_strbytes(1))
  Conv_bytes(3) = Hexval(conv_strbytes(2))
  Conv_bytes(2) = Hexval(conv_strbytes(3))
  Conv_bytes(1) = Hexval(conv_strbytes(4))
End Sub

' **
' * Used to convert a long value to 4 bytes
' **
Sub Convertword(byval Wordvalue As Word)

  'set the long value outside the frame
  Convw_word = Wordvalue
  Convw_string = Hex(convw_word)

  'parse the HEX string value as 2 char strings
  Convw_strbytes(1) = Mid(convw_string , 1 , 2)
  Convw_strbytes(2) = Mid(convw_string , 3 , 2)

  'convert the bit order to little endian
  Convw_bytes(2) = Hexval(convw_strbytes(1))
  Convw_bytes(1) = Hexval(convw_strbytes(2))
End Sub

' **
' * This sub sets the debug output device
' *
' * Set to 1 for RS232
' * Set to 2 for LCD
' **
Sub Setdebug(byval Debugvalue As Byte)

  'set the global debug value
  G_debug = Debugvalue

End Sub


' **
' * This sub prints a message to the RS232 or LCD if the debug is set
' **
Sub Debug(byval Debugtext As String)

  'if debug is set to 1, select the RS232 output device
  If G_debug = 1 Then
    Print Debugtext
  End If

  'if debug is set to 2, select the LCD output device
  If G_debug = 2 Then
    Lcd Debugtext
  End If

End Sub

' **
' * This function is used to parse a 512 byte sector for the 32 byte fat 
record
' **
Function Parsefatrecord(byval Recordnum As Word) As Byte

  'start at byte 0 offset
  G_offset = Recordnum * 32

  G_parse_temp = G_offset + 12

  'set the offset and read the attribute byte
  G_attribute = G_xferbuffer(g_parse_temp)

  'Print "REC: " ; Recordnum ; " OFFSET: " ; Offset ; "  ATTRIB: " ; 
Attribute

  'For Cntr = 1 To 32
  '  Parse_temp = Offset + Cntr
  '  Print Hex(xferbuffer(parse_temp)) ; " " ;
  'Next Cntr

  'Print ""

  'For Cntr = 1 To 32
  '  Parse_temp = Offset + Cntr
  '  Print Xferbuffer(parse_temp) ; " " ;
  'Next Cntr

  'Print ""

  'check the attribute byte to determine if this a a LFN ercord.
  'The attribute indicates that this is a LFN record if it's OFh
  If G_attribute = &H0F Then

    'LFN code goes here

  Else
    Fat_attribute = G_attribute

    'Print "VALUES: " ;
    'get the 8 byte short file name
    For G_cntr = 1 To 8
      G_offset = G_offset + 1
      Fat_shortname(g_cntr) = G_xferbuffer(g_offset)
    Next G_cntr

    'get the 3 byte file extension
    For G_cntr = 1 To 3
      G_offset = G_offset + 1
      Fat_extension(g_cntr) = G_xferbuffer(g_offset)
    Next G_cntr

    'get the files created milliseconds (ms)
    G_offset = G_offset + 1                                 'base 1 array 
requires +1
    Fat_createdms = G_xferbuffer(g_offset + 1)

    'get the create time as a word
    G_offset = G_offset + 2
    Fat_createdtm = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the created date as a word
    G_offset = G_offset + 2
    Fat_createddt = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the last access date as a word
    G_offset = G_offset + 2
    Fat_accessdt = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the high 16 bits of the fat cluster as a word
    G_offset = G_offset + 2
    Fat_fathighcluster = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the last update time as a word
    G_offset = G_offset + 2
    Fat_updatetime = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the last update date as a word
    G_offset = G_offset + 2
    Fat_updatedate = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the 16 bit fat cluster as a word
    G_offset = G_offset + 2
    Fat_cluster = Getwordfrombuffer(g_xferbuffer(1) , G_offset)

    'get the file size as a dword (long)
    G_offset = G_offset + 2
    'Print "long offset = " ; Offset
    Fat_filesize = Getlongfrombuffer(g_xferbuffer(1) , G_offset)

    If Fat_attribute > 31 And Fat_shortname(1) <> 229 Then
       For G_temp_cntr = 1 To 4
          G_temp_val = G_temp_cntr + G_offset
          'Print Xferbuffer(temp_val) ; " " ;
       Next G_temp_cntr
       'Print " = " ; Fat_filesize ; " Bytes at offset " ; Offset ; " cntr=" 
; Temp_cntr
    End If

    'point the offset to the end of the 32 byte record
    G_offset = G_offset + 4

    'set the volume name of attribute indicates this is a file
    If Fat_attribute > 31 And Fat_shortname(1) <> 229 Then
      Fat_filename = ""
        For G_cntr = 1 To 8                                 '
         If Fat_shortname(g_cntr) < 129 And Fat_shortname(g_cntr) > 33 Then
           'Print "'" ; Chr(fat_shortname(cntr)) ; "' " ; 
Hex(fat_shortname(cntr)) ; " " ;
           Fat_filename = Fat_filename + Chr(fat_shortname(g_cntr))
         End If
       Next G_cntr
       Fat_filename = Fat_filename + "."
       For G_cntr = 1 To 3
         If Fat_extension(g_cntr) < 129 And Fat_extension(g_cntr) > 33 Then
          Fat_filename = Fat_filename + Chr(fat_extension(g_cntr))
         End If
       Next Cntr
          'Print "file: " ; Fat_filename
    End If

  End If

  Parsefatrecord = G_attribute

End Function

' **
' * This function is used to list the files on the root dir
' **
Sub Listrootdir()

  'empty return value
  Local Returnvalue As Byte

  'call 'search' with a bogus filename
  ' NOTE: wild cards '*' are meaningless, this is a totally bogus filename
  Returnvalue = Searchrootdir( "*.*")

End Sub

' **
' * Lists a file contents as ascii text
' **
Sub Listfile(byval Fname As String)

  'search results
  Local Result As Byte

  'search for file
  Result = Searchrootdir(fname)

  'if file was found, dump it as ascii text
  If Result > 0 Then

    'dump the contents
    Call Dumpfile(fat_cluster , Fat_filesize)

  End If

End Sub

' **
' * This function is used to search the root dir for a file
' **
Function Searchrootdir(byval Searchname As String) As Byte

  'set the value out of the frame
  G_touchholder = Searchname

  'point to the first byte of the array
  G_pointer = Varptr(g_xferbuffer(1))

  Print "FAT start sector:" ; Br_actualstartsector ; " for dir entries"
  'Print "data start sector:" ; Datasector

  Print "Searching for " ; G_touchholder

  While G_rootloop <= Br_sectorsinroot

     'read sector number 1 into the 512 byte array starting at the pointer 
address
     Readsector G_pointer , Br_actualstartsector , 1

     'set the fat directory sector for future updates
     Fat_dirsector = Br_actualstartsector

     'start with record 0
     G_volumerecord = 0

     'exame each of 16 records on this sector until the volume attribute is 
found
     While G_volumerecord < 16

       'set the fat directory entry number for future updates
       Fat_direntry = G_volumerecord

       G_attribreturn = Parsefatrecord(g_volumerecord)

         'the attribute indicates this entry is a file
         If G_attribreturn > 31 And Fat_shortname(1) <> 229 Then

           Print "File: " ; Fat_filename

           'found the filename so return
           If Fat_filename = G_touchholder Then

             'set the return value and exit
             Searchrootdir = G_attribreturn
             G_rootloop = 99
             G_volumerecord = 99
           Else
             'file not found
             Searchrootdir = 0
           End If

         End If

       G_volumerecord = G_volumerecord + 1
     Wend

     Br_actualstartsector = Br_actualstartsector + 1
     G_rootloop = G_rootloop + 1
  Wend

End Function

' **
' * copy Memory from (Z) nach (X)
' * counts of bytes in r24
' *
' * Code originated by Josef Franz
' **
_mem_copy:
   ld r25, z+
   st x+, r25
   dec r24
   brne _Mem_Copy
   ret

